declare type Func = (currentValue: any, index?: number) => any
declare type Predicate = (currentValue: any, index?: number) => boolean
declare type Accumulator = (accumulatedValue: any, currentValue: any, index?: number) => any

declare var XArray : {
    fromSize: (size?: number) => XArray<any>;
    fromRange: (from: number, to: number) => XArray<any>;
}

declare interface XArray<T> extends Array<T> {
    new<T>(...items) : XArray<T>;

    toArray: () => any[];
    distinct: () => XArray<T>;
    accumulate: (initial: any, accum?: Accumulator) => any;
    union: (array: Array<any>) => XArray<T>;

    select: (func?: Func) => XArray<T>;
    where: (func?: Predicate) => XArray<T>;

    orderBy: (func?: Func) => XArray<T>;
    orderByDesc: (func: Func) => XArray<T>;

    firstOrDefault: (predicate?: Predicate) => T;
    first: (predicate?: Predicate) => T;

    lastOrDefault: (predicate?: Predicate) => T;
    last: (predicate?: Predicate) => T;

    all: (predicate?: Predicate) => boolean;
    any: (predicate?: Predicate) => boolean;

    take: (count: number) => XArray<T>;
    skip: (count: number) => XArray<T>;

    sum: (func?: Func) => number;
    avg: (func?: Func) => number;
    count: (func?: Func) => number;
    min: (func?: Func) => number;
    max: (func?: Func) => number;

    /* 
     * Groups the array values into a dictionary of arrays. 
     * Grouping keys are determined by applying the <func> function arg to each element 
     */
    groupBy: (func) => any;

    /* 
     * Groups this array into an dictionary of <keyCount> arrays,
     * each containing the same number of elements, except possibly the last array, 
     * which may have less elements than the others 
     * if the length of this array is not divisible by <keyCount>
     */
    groupByNumber(keyCount: number): any;

    /*
     * groups this array according to <key> generator function and then 
     * aggregates the resulting dictionary over <value> generator function
     */
    groupAggregate(key, value, aggregation): any;

    removeItem: (item: any) => XArray<T>;
    removeAt: (index: number, count?: number) => XArray<T>;
    insert: (index: number, ...items: any[]) => XArray<T>;
    append: (items: any[]) => XArray<T>;

    flatten: () => XArray<T>;
}